; *****************************************************************************
; **                 MASM32 Library Search Utility v0.9 by                   ** 
; **                Mark Jones (gzscuqn02ATsneakemailD0Tcom)                 **
; **                                                                         **
; **    Enumerates all files in \masm32\include and \masm32\macros for a     **  
; **    specified text string. Useful for finding which includes have the    **
; **   prototypes needed to build a project or where a particular macro is   **
; **   located. Up to five search paths are customizable and settings are    **
; **  saved in an .ini file.                             (c) MCJ 2006.03.23  **
; *****************************************************************************

include IncFind.inc                                 ; all includes for this app

.code
start:
    invoke GetModuleHandle,0                        ; get instance handle
    mov hInstance,eax

    invoke InitCommonControls                       ; needed for XP theme
    invoke DialogBoxParam,hInstance,101,0,addr DlgProc,0
    invoke ExitProcess,0


DlgProc proc hWin:HWND,uMsg:UINT,wParam:WPARAM,lParam:LPARAM
    local buffer[256]:byte
    mov eax,uMsg
    
    .if eax==WM_INITDIALOG
        m2m hWnd,hWin                               ; save hWin as hWnd
        m2m OFN.hwndOwner,hWnd                      ; set openfilename presets
        mov OFN.lStructSize,sizeof OFN              ; 
        mov OFN.lpstrFilter,offset ofnFilter        ; file filter
        mov OFN.lpstrFile,offset ofnPath            ; file path
        mov OFN.lpstrTitle,offset ofnTitle          ; title
        mov OFN.nMaxFile,255                        ; reserve a null

        invoke SetWindowText,hWnd,addr szTitle      ; title main window
        invoke GetMenu,hWnd                         ; get our menu handle
        mov hMenu,eax                               ;
        mov nCase,1                                 ; default to case-insensitive
        invoke GlobalAlloc,GPTR,65536               ; 64kb should be enough for list
        mov hFiles,eax                              ; save "files" pointer
        call ReadINI                                ; read .ini file
        call EnumPaths                              ; build file list

        invoke GetDlgItem,hWnd,1010                 ; get handle of listview
        mov hLV,eax                                 ; control as hLV
        mov lvi.imask,LVIF_TEXT                     ; preset item info
        mov lvc.imask,LVCF_TEXT+LVCF_WIDTH
        mov lvc.pszText,offset szlvc
        mov lvc.lx,174
        invoke SendMessage,hLV,LVM_INSERTCOLUMN,1,addr lvc
        invoke GetSysColor,COLOR_WINDOWTEXT         ; set listview colors
        invoke SendMessage,hLV,LVM_SETTEXTCOLOR,0,eax
        invoke GetSysColor,COLOR_BTNFACE            ; to match window's
        push eax
        invoke SendMessage,hLV,LVM_SETBKCOLOR,0,eax
        pop eax
        invoke SendMessage,hLV,LVM_SETTEXTBKCOLOR,0,eax
        
    .elseif eax==WM_COMMAND
        mov eax,wParam                              ; wParam == (condition):(control)
        mov ecx,eax                                 ; ax = control id
        shl ecx,16                                  ; cx = condition
        .if ax==1003 && cx==BN_CLICKED              ; ID: find button AND clicked?
            push esi                                ; clear listview loop
            mov lvi.iItem,0                         ; point at first item
            invoke SendMessage,hLV,LVM_DELETEALLITEMS,esi,0
            call DoSearch                           ; perform search
        .elseif ax==1009 && cx==BN_CLICKED          ; edit button?
            call EditFile
        .elseif ax==10004                           ; launch options dialog as child
            invoke DialogBoxParam,hInstance,1100,hWnd,addr OptionsDlgProc,0
        .elseif ax==10007                           ; index menu
            call EnumPaths 
        .elseif ax==10003                           ; exit menu
            invoke SendMessage,hWnd,WM_CLOSE,0,0
        .elseif ax==10006                           ; case sensitivty toggle
            .if nCase == 0
                invoke ModifyMenu,hMenu,10006,0,10006,addr szCaseInSens
                mov nCase,1
            .else
                invoke ModifyMenu,hMenu,10006,0,10006,addr szCaseSens
                mov nCase,0
            .endif
            invoke DrawMenuBar,hWnd                 ; update menu bar
        .elseif ax==10005                           ; about menu
            invoke ShellAbout,hWnd,addr szTitle,addr szAboutMsg,0
        .endif

    .elseif uMsg==WM_NOTIFY                         ; listview feedback
        push edi
        mov edi,lParam
        assume edi:ptr NMHDR
        mov eax,[edi].hwndFrom
        .if eax==hLV                                ; is it the listview?
            .if [edi].code==NM_DBLCLK               ; was item double-clicked?
	            invoke SendMessage,hLV,LVM_GETNEXTITEM,-1,LVNI_FOCUSED
	            mov lvi.iItem,eax
	            mov lvi.iSubItem,0
	            mov lvi.imask,LVIF_TEXT
	            lea eax,buffer
	            mov lvi.pszText,eax
	            mov lvi.cchTextMax,256
	            invoke SendMessage,hLV,LVM_GETITEM,0,addr lvi
	            invoke EditFile,addr buffer         ; open the file
            .endif
        .endif
    
    .elseif eax==WM_CLOSE
        invoke GlobalFree,hFiles                    ; free file buffer
        invoke EndDialog,hWnd,0
    .elseif eax==WM_DESTROY
        invoke PostQuitMessage,0
        
    .else
        mov eax,0                                   ; message not handled flag,
        ret                                         ; pass to default msg handler
    .endif
    mov eax,1                                       ; messsage handled flag, do
    ret                                             ; not pass to default msg hdlr
DlgProc endp



OptionsDlgProc proc hWinO:HWND,uMsgO:UINT,wParamO:WPARAM,lParamO:LPARAM
    mov eax,uMsgO
    
    .if eax==WM_INITDIALOG
        m2m hWndO,hWinO
        SetOptionsSZ                                ; populate paths from buffers

    .elseif eax==WM_COMMAND
        mov eax,wParamO
        mov ecx,eax
        shl ecx,16
        .if ax==1111 && cx==BN_CLICKED              ; path1 browse button
            invoke GetNewPath,addr path1
        .elseif ax==1112 && cx==BN_CLICKED          ; path2 browse button
            invoke GetNewPath,addr path2
        .elseif ax==1113 && cx==BN_CLICKED
            invoke GetNewPath,addr path3
        .elseif ax==1114 && cx==BN_CLICKED
            invoke GetNewPath,addr path4
        .elseif ax==1115 && cx==BN_CLICKED
            invoke GetNewPath,addr path5
        .elseif ax==1116 && cx==BN_CLICKED          ; editor browse button 
            invoke GetNewPath,addr path6
        .elseif ax==1102 && cx==BN_CLICKED          ; ok button, save changes
            GetOptionsSZ                            ; update string buffers
            call SaveINI                            ; save new fields to .ini file
            call EnumPaths                          ; preload new dirs
            invoke EndDialog,hWinO,0                ; and close this window
        .elseif ax==1103 && cx==BN_CLICKED          ; cancel button, discard changes
            invoke EndDialog,hWinO,0
        .endif
        
    .elseif eax==WM_CLOSE || eax==WM_DESTROY        ; only EndDialog, don't PostQuitMsg
        invoke EndDialog,hWinO,0
        
    .else
        mov eax,0
        ret
    .endif
    mov eax,1
    ret
OptionsDlgProc endp



ReadINI proc uses esi edi
    invoke GetModuleFileName,0,addr SelfDir,255     ; .\IncFind.exe
    mov ecx,offset SelfDir                          ; ecx = pointer
@@:
    mov al,byte ptr[ecx]                            ; find end
    inc ecx
    test al,al
    jnz @B
    mov dword ptr[ecx-4],00696E69h                  ; make extension ".ini"
    invoke GetFileAttributes,addr SelfDir
    .if eax!=(-1)                                   ; if .ini file exists
        call LoadINI                                ; read locations
    .else                                           ; if no .ini file, revert to 
        invoke szCopy,addr path1,addr szInclude     ; default search location
        call SaveINI                                ; save .ini file now
    .endif

    invoke szCopy,addr szTemp,addr szSearching      ; set statusbar text
    invoke lstrcat,addr szTemp,addr path1           ; "Searching " + path1,
    invoke SetDlgItemText,hWnd,1006,addr szTemp     ;
    ret
ReadINI endp



GetNewPath proc uses esi edi szPath:DWORD
    invoke GetOpenFileName,offset OFN               ; ask for file
    .if eax!=0                                      ; if a file is returned,
        mov esi,offset ofnPath                      ; setup text copy source
        mov edi,szPath                              ; and dest
@@:                                                 ; 
        mov al,byte ptr [esi]                       ; copy source byte into al
        mov byte ptr [edi],al                       ; copy byte to destination
        inc esi                                     ; increment source and
        inc edi                                     ; destination pointers
        test al,al                                  ; null yet?
        jnz @B                                      ; loop until all copied
@@:
        dec esi                                     ; now go backwards in src
        mov al,byte ptr [esi]                       ;
        cmp al,"."                                  ; to last "."
        jnz @B
        push [esi]                                  ; save file extension and "."
@@:
        mov al,byte ptr [edi]                       ; fetch destination byte
        dec edi                                     ; search backwards from end
        cmp al,"\"                                  ; look for the last "\"
        jnz @B                                      ; loop until there
        pop eax                                     ; restore file extension as eax
        mov byte ptr [edi+2],"*"                    ; add an asterisk to dest
        mov dword ptr [edi+3],eax                   ; and the 4-byte extension
        mov byte ptr [edi+7],00                     ; plus a terminating null
        invoke SendMessage,hWndO,WM_INITDIALOG,0,0  ; update display
    .endif
    ret
GetNewPath endp



SaveINI proc
    invoke WritePrivateProfileString,addr szAppName,addr szKey1,addr path1,addr SelfDir
    invoke WritePrivateProfileString,addr szAppName,addr szKey2,addr path2,addr SelfDir
    invoke WritePrivateProfileString,addr szAppName,addr szKey3,addr path3,addr SelfDir
    invoke WritePrivateProfileString,addr szAppName,addr szKey4,addr path4,addr SelfDir
    invoke WritePrivateProfileString,addr szAppName,addr szKey5,addr path5,addr SelfDir
    invoke WritePrivateProfileString,addr szAppName,addr szKey6,addr path6,addr SelfDir
    ret
SaveINI endp



LoadINI proc
    invoke GetPrivateProfileString,addr szAppName,addr szKey1,0,addr path1,255,addr SelfDir
    invoke GetPrivateProfileString,addr szAppName,addr szKey2,0,addr path2,255,addr SelfDir
    invoke GetPrivateProfileString,addr szAppName,addr szKey3,0,addr path3,255,addr SelfDir
    invoke GetPrivateProfileString,addr szAppName,addr szKey4,0,addr path4,255,addr SelfDir
    invoke GetPrivateProfileString,addr szAppName,addr szKey5,0,addr path5,255,addr SelfDir
    invoke GetPrivateProfileString,addr szAppName,addr szKey6,0,addr path6,255,addr SelfDir
    ret
LoadINI endp



EditFile proc lpFile:DWORD
    cmp byte ptr [path6],00         ; is an editor defined? 
    jz defaultEd
    invoke ShellExecute,hWnd,0,addr path6,lpFile,0,SW_SHOW
    ret
defaultEd:                          ; if not use file association
    invoke ShellExecute,hWnd,0,lpFile,0,0,SW_SHOW
    ret
EditFile endp


    ; fast general-purpose string copy routine by Mark Jones 2005
align 16
szCopy proc uses esi edi szDest:DWORD,szSource:DWORD
    mov esi,szSource
    cld                             ; copy forwards
    mov edi,szDest
@@:
    cmp byte ptr [esi+03],00        ; better branching
    jz CopyFour
    cmp byte ptr [esi+02],00
    jz CopyThree
    cmp byte ptr [esi+01],00
    jz CopyTwo
    cmp byte ptr [esi+00],00
    jz CopyOne
    movsd                           ; else move DW ESI --> EDI & increment
    jmp @B
CopyFour:                           ; finish remaining bytes
    movsd
    ret
CopyTwo:
    movsw
    ret
CopyThree:
    movsw
CopyOne:
    movsb
    ret
szCopy endp



EnumPaths  proc  uses ebx edi esi
    mov edi,[hFiles]                                ; edi = mem destination offset

; enumerate all matching files in all paths
    lea ebx,path1                                   ; load path offsets
    call enum                                       ; enumerate
    lea ebx,path2
    call enum
    lea ebx,path3
    call enum
    lea ebx,path4
    call enum
    lea ebx,path5
    call enum
    mov byte ptr [edi],00                           ; terminate list with a null
;invoke MessageBox,0,dword ptr [hFiles],0,MB_OK ; debugging
    ret

enum:
    invoke FindFirstFile,ebx,offset FindData        ; begin folder itemization
    .if eax==INVALID_HANDLE_VALUE                   ; if no files found,
        retn                                        ; exit without error
    .else                                           ; first term needs path
        mov hFind,eax                               ; save search handle
@@:
        mov al,byte ptr [ebx]                       ; fetch path byte
        inc ebx                                     ; copy forward
        mov byte ptr[edi],al                        ; else copy byte
        inc edi                                     ; and increment
        cmp al,"*"                                  ; up to the *
        jnz @B                                      ; then fall through
    .endif

; create null-delimited list of all files from each search path
FindAnother:                                        ; loop until all files found
    mov esi,offset FindData.cFileName               ; esi = filename pointer
@@:                                                 ; copy found filename into buffer
    mov al,byte ptr [esi]                           ; fetch source byte
    inc esi                                         ; increment pointers
    mov byte ptr [edi],al                           ; put into dest
    inc edi
    test al,al                                      ; was byte a null?
    jnz @B                                          ; else loop
;mov byte ptr [edi-1],","    ; replace null with comma for debugging

    invoke FindNextFile,hFind,offset FindData       ; find another file
    test al,al                                      ; if no more files exist,
    jz Continue                                     ; start searching through results
    jmp FindAnother                                 ; else keep enumerating files

Continue:
    invoke FindClose,hFind
    retn                                            ; near return only!
EnumPaths endp



SearchLib  proc  uses edi esi
    local path[256]:BYTE

    mov esi,hFiles                                  ; esi = list of files to search
; loop here, open and search each file sequentially
Search:
; first item in list will include a path, use that for following files
    cmp byte ptr [esi+00],00                        ; if null,
    jz Skip                                         ; we're at the end
    cmp byte ptr [esi+00],"\"                       ; if \
    jz updatepath                                   ; treat it as a path
    cmp byte ptr [esi+01],":"                       ; colon too...
    jz updatepath                                   ;
    jmp fileopen                                    ; no path, search file
    
updatepath:
    lea edi,path                                    ; path buffer as edi
@@:
    mov al,byte ptr [esi]                           ; fetch path byte
    inc esi
    cmp al,"*"                                      ; break when "*" found
    jz @F
    mov byte ptr [edi],al                           ; copy path
    inc edi
    jmp @B
@@:                                                 ; esi now == first file
    mov byte ptr [edi],00                           ; path == path
    
; open the path + file
fileopen:
    invoke szCopy,addr szFile,addr path
    invoke lstrcat,addr szFile,esi

    invoke CreateFile,addr szFile,GENERIC_READ,FILE_SHARE_READ,0,\
        OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL or FILE_FLAG_SEQUENTIAL_SCAN,0
    mov hSearch,eax                                 ; always save handle
    .if eax==INVALID_HANDLE_VALUE                   ; if file cannot be opened
        invoke Beep,200,500                         ; error sound
    .endif

; how big is it?    
    invoke GetFileSize,hSearch,0                    ; get size of opened file
    mov nFileSize,eax
    .if eax==(-1)
        invoke Beep,300,500                         ; size of file not reported!
        mov nFileSize,0
    .endif
    
; allocate a ram buffer 1 byte larger than file
    inc nFileSize                                   ; allocate ram+1 for file
    invoke GlobalAlloc,GPTR,nFileSize
    mov hBuffer,eax

; indicate which file we're about to search
    invoke SetDlgItemText,hWnd,1006,addr szFile     ; set window status text

; load file into ram
    invoke ReadFile,hSearch,hBuffer,nFileSize,addr nReadBytes,0
    .if eax==0
        invoke Beep,400,500                         ; error reading file
    .endif

; search for term in ram (case sensitive) and update master results
    mov ecx,dword ptr hBuffer                       ; ecx = file buffer pointer
    cmp byte ptr[nCase],1                           ; we searching case-sensitive?
    jz NoCase
; case-sensitive version
RestartQuery:                                       ; land here if match failed
    mov edx,offset szSearchQuery                    ; edx = search term pointer
    mov ah,byte ptr [edx]
align 8
@@:                                                 ; 
    mov al,byte ptr [ecx]                           ; get byte into al
    inc ecx                                         ;
    test al,al                                      ; are we at EOF?
    jz EOF                                          ; 
    cmp al,ah                                       ; does it match query byte?
    jnz @B                                          ; if not, loop source
align 8
@@:                                                 ; first letter matches
    inc edx                                         ; increment query byte
    mov al,byte ptr [ecx]                           ; get next file byte
    cmp byte ptr [edx],00                           ; did we match whole string?
    jz Match                                        ; we did, cool!
    test al,al                                      ; at source EOF?
    jz EOF                                          ; 
    inc ecx                                         ; else keep searching
    cmp al,byte ptr [edx]                           ; still matching 
    jz @B                                           ; 
    dec ecx                                         ; if no match,
    jmp RestartQuery                                ; reset and keep looking
Match:                                              ; term found!
    mov lvi.pszText,offset szFile                   ; create listview entry
    invoke SendMessage,hLV,LVM_INSERTITEM,0,addr lvi
    inc lvi.iItem                                   ; and increment
EOF:                                                ; else term not found
    jmp SearchComplete

; case-insensitive version (41h-5Ah = caps, 61h-7Ah = smalls)
NoCase:                                             ; land here if match failed
    mov edx,offset szSearchQuery                    ; edx = search term pointer
    mov ah,byte ptr [edx]                           ; ah = first byte of term
align 8
@@:                                                 ; 
    mov al,byte ptr [ecx]                           ; get byte into al
    inc ecx                                         ;
    test al,al                                      ; are we at EOF?
    jz EOF                                          ; 
    cmp al,ah                                       ; does it match query byte?
    jz @F                                           ; if so, keep matching
    cmp al,"A"
    jl @B                                           ; reject non-ascii leters
    cmp al,"z"
    jg @B
    cmp al,"Z"
    jle convupper
    cmp al,"a"
    jge convlower
    jmp @B                                          ; reject anything else
convlower:
    sub al,32
    jmp tryitagain
convupper:
    add al,32
tryitagain:
    mov al,byte ptr [ecx]                           ; get byte into al
    cmp al,ah                                       ; does it match query byte?
    jz @F                                           ; if so, keep matching
align 8
@@:                                                 ; first letter matches
    inc edx                                         ; increment query byte
    mov al,byte ptr [ecx]                           ; get next file byte
    cmp byte ptr [edx],00                           ; did we match whole string?
    jz cMatch                                       ; we did, cool!
    test al,al                                      ; at source EOF?
    jz cEOF                                         ; not found
    inc ecx                                         ; else keep searching
    cmp al,byte ptr [edx]                           ; still matching 
    jz @B                                           ; loop if exact 
    cmp al,"A"
    jl @F                                           ; reject non-ascii leters
    cmp al,"z"
    jg @F
    cmp al,"Z"
    jle upper
    cmp al,"a"
    jge lower
    jmp @F                                          ; reject anything else
lower:
    sub al,32                                       ; convert upper to lower
    jmp tryagain
upper:
    add al,32                                       ; lower to upper
tryagain:
    cmp al,byte ptr [edx]                           ; does it match now? 
    jz @B                                           ; if not, no match
align 8
@@:
    dec ecx                                         ; if no match,
    jmp NoCase                                      ; reset and keep looking
cMatch:                                             ; term found!
    mov lvi.pszText,offset szFile
    invoke SendMessage,hLV,LVM_INSERTITEM,0,addr lvi
    inc lvi.iItem
cEOF:                                               ; term not found

SearchComplete:
; clean up and check wether any more files are listed
    invoke GlobalFree,hBuffer                       ; close file mem buffer
    invoke CloseHandle,hSearch                      ; close file handle
@@:
    mov al,byte ptr [esi]                           ; move file pointer to next
    inc esi                                         ; file
    test al,al                                      ; entry
    jnz @B                                          ;
    jmp Search                                      ; loop until all files are searched

; all files searched, we made it!
    invoke FindClose,hFind                          ; free resources
Skip:
    ret
SearchLib  endp



DoSearch proc
    invoke GetDlgItemText,hWnd,1002,addr szSearchQuery,255 ; get search term
    invoke GetTickCount                             ; let's time our search
    push eax                                        ; save current uptime
    
    invoke SearchLib                                ; perform full search

    invoke GetTickCount                             ; get new time in eax
    pop ecx                                         ; restore old time as ecx
    sub eax,ecx                                     ; subtract new from old
    invoke dwtoa,eax,addr szTime                    ; convert to string
    invoke szCopy,addr szTemp,addr szSearched       ; " Searched in "
    invoke lstrcat,addr szTemp,addr szTime          ; "734"
    invoke lstrcat,addr szTemp,addr ms              ; "ms"
    invoke SetDlgItemText,hWnd,1006,addr szTemp     ; display done message
    ret
DoSearch endp


end start
